package org.ogce.xregistry.client;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.List;

import javax.xml.namespace.QName;

import org.ietf.jgss.GSSCredential;
import org.ogce.xregistry.security.CertificateManager;
import org.ogce.xregistry.utils.XRegClientConstants;
import org.ogce.xregistry.utils.XRegistryClientException;

import xregistry.generated.AddCapabilityDocument;
import xregistry.generated.AddCapabilityDocument.AddCapability;
import xregistry.generated.AddOGCEResourceDocument;
import xregistry.generated.AddOGCEResourceDocument.AddOGCEResource;
import xregistry.generated.AddResourceDocument;
import xregistry.generated.AddResourceDocument.AddResource;
import xregistry.generated.App2HostsDocument;
import xregistry.generated.App2HostsResponseDocument;
import xregistry.generated.CapabilityToken;
import xregistry.generated.FindAppDescDocument;
import xregistry.generated.FindAppDescResponseDocument;
import xregistry.generated.FindAppDescResponseDocument.FindAppDescResponse.AppData;
import xregistry.generated.FindHostsDocument;
import xregistry.generated.FindHostsResponseDocument;
import xregistry.generated.FindOGCEResourceDocument;
import xregistry.generated.FindOGCEResourceDocument.FindOGCEResource;
import xregistry.generated.FindOGCEResourceResponseDocument;
import xregistry.generated.FindResourceDocument;
import xregistry.generated.FindResourceResponseDocument;
import xregistry.generated.FindServiceDescDocument;
import xregistry.generated.FindServiceDescResponseDocument;
import xregistry.generated.FindServiceInstanceDocument;
import xregistry.generated.FindServiceInstanceResponseDocument;
import xregistry.generated.GetAbstractWsdlDocument;
import xregistry.generated.GetAbstractWsdlResponseDocument;
import xregistry.generated.GetAppDescDocument;
import xregistry.generated.GetAppDescDocument.GetAppDesc;
import xregistry.generated.GetCapabilityDocument;
import xregistry.generated.GetCapabilityDocument.GetCapability;
import xregistry.generated.GetCapabilityResponseDocument;
import xregistry.generated.GetConcreateWsdlDocument;
import xregistry.generated.GetHostDescDocument;
import xregistry.generated.GetOGCEResourceDocument;
import xregistry.generated.GetOGCEResourceDocument.GetOGCEResource;
import xregistry.generated.GetResourceDocument;
import xregistry.generated.GetServiceDescDocument;
import xregistry.generated.HostDescData;
import xregistry.generated.IXregistryPortType;
import xregistry.generated.IsAuthorizedToAcssesDocument;
import xregistry.generated.IsAuthorizedToAcssesDocument.IsAuthorizedToAcsses;
import xregistry.generated.IsAuthorizedToAcssesResponseDocument;
import xregistry.generated.OGCEResourceData;
import xregistry.generated.RegisterAppDescDocument;
import xregistry.generated.RegisterConcreteWsdlDocument;
import xregistry.generated.RegisterConcreteWsdlDocument.RegisterConcreteWsdl;
import xregistry.generated.RegisterHostDescDocument;
import xregistry.generated.RegisterServiceDescDocument;
import xregistry.generated.RegisterServiceDescDocument.RegisterServiceDesc;
import xregistry.generated.RemoveAppDescDocument;
import xregistry.generated.RemoveAppDescDocument.RemoveAppDesc;
import xregistry.generated.RemoveCapabilityDocument;
import xregistry.generated.RemoveCapabilityDocument.RemoveCapability;
import xregistry.generated.RemoveConcreteWsdlDocument;
import xregistry.generated.RemoveHostDescDocument;
import xregistry.generated.RemoveOGCEResourceDocument;
import xregistry.generated.RemoveOGCEResourceDocument.RemoveOGCEResource;
import xregistry.generated.RemoveResourceDocument;
import xregistry.generated.RemoveServiceDescDocument;
import xregistry.generated.ResourceData;
import xregistry.generated.ServiceDescData;
import xregistry.generated.WsdlData;
import xsul.MLogger;

public class XRegistryClient {
	  protected static MLogger log = MLogger.getLogger(XRegClientConstants.LOGGER_NAME);
	    
	  //Stub object to be initialized in the constructor and reused within the class.
	  private IXregistryPortType xregistryStub;
	  
	/**
	 * Constructs a XRegistryClient for non-secure XRegistry Operations.
	 * @param non-secured xregistry url
	 */
	  public XRegistryClient(String nonSecureXRegistryUrl) throws XRegistryClientException{
		  //if the service wsdl end point is secure then a secure connection has to be created 
		  //by passing security credentials. 
		  if (nonSecureXRegistryUrl.startsWith("https")) {
			  throw new XRegistryClientException("This is a secure service, security information has to be passed" +
			  		"Only non-secured services can be connected by passing only epr.");
		  }
		  //Double Check to make sure the service end point has ?wsdl needed for WSIF
		  if (!nonSecureXRegistryUrl.endsWith("?wsdl")) {
			 nonSecureXRegistryUrl += "?wsdl";
		  }
		  this.xregistryStub = XRegistryConnection.createNonSecureServiceStub(nonSecureXRegistryUrl);
		  log.info("Created a non Secure XRegistry Connection Stub");
	  }
	  
		/**
		 * Constructs a XRegistryClient with GSICredetial for secure XRegistry Operations .
		 * @param gsicredential
		 * @param trusted certificates
		 * @param secured-xregistry-url
		 * 
		 */
	    public XRegistryClient(GSSCredential gssCredential,  X509Certificate[] trustedCerts,
		           String xregistryServiceWsdlUrl) throws XRegistryClientException{
			  //Double Check to make sure the service end point has ?wsdl needed for WSIF
			  if (!xregistryServiceWsdlUrl.endsWith("?wsdl")) {
				  xregistryServiceWsdlUrl += "?wsdl";
			  }
			  
			  //if the service is a non-secured service, then no credentials are required to connect. 
			  if (!xregistryServiceWsdlUrl.startsWith("https")) {
				  this.xregistryStub = XRegistryConnection.createNonSecureServiceStub(xregistryServiceWsdlUrl);
				  log.info("Created a non Secure XRegistry Connection Stub");
			  } else {
				  this.xregistryStub =XRegistryConnection.createGSISecureServiceStub(gssCredential, trustedCerts, xregistryServiceWsdlUrl);
				  log.info("Created a GSI Credential Secured XRegistry Connection Stub");
			  }
		 }
	    
	    /**
		 * Constructs a XRegistryClient with GSICredetial and trusted ca's file for secure XRegistry Operations .
		 * @param gsicredential
		 * @param trusted certificates
		 * @param secured-xregistry-url
		 * 
		 */
	    public XRegistryClient(GSSCredential gssCredential,  String trustedCerts,
		           String xregistryServiceWsdlUrl) throws XRegistryClientException{
	    	this(gssCredential, CertificateManager.getTrustedCertificates(trustedCerts), xregistryServiceWsdlUrl);
		 }
	    
		/**
		 * Constructs a XRegistryClient with SSL Certificate File for secure XRegistry Operations .
		 * @param gsicredential
		 * @param trusted certificates
		 * @param secured-xregistry-url
		 * 
		 */
	    public XRegistryClient(String hostkeyPath,  X509Certificate[] trustedCerts,
		           String xregistryServiceWsdlUrl) throws XRegistryClientException{
			  //Double Check to make sure the service end point has ?wsdl needed for WSIF
			  if (!xregistryServiceWsdlUrl.endsWith("?wsdl")) {
				  xregistryServiceWsdlUrl += "?wsdl";
			  }
			  
			  //if the service is a non-secured service, then no credentials are required to connect. 
			  if (!xregistryServiceWsdlUrl.startsWith("https")) {
				  this.xregistryStub = XRegistryConnection.createNonSecureServiceStub(xregistryServiceWsdlUrl);
				  log.info("Created a non Secure XRegistry Connection Stub");
			  } else {
				  try {	  
					this.xregistryStub =XRegistryConnection.createSSLSecureServiceStub(hostkeyPath, trustedCerts, xregistryServiceWsdlUrl);
					log.info("Created a SSL Credential Secured XRegistry Connection Stub");
				} catch (FileNotFoundException e) {
					throw new XRegistryClientException(e.getLocalizedMessage(),e);
				} catch (GeneralSecurityException e) {
					throw new XRegistryClientException(e.getLocalizedMessage(),e);
				} catch (IOException e) {
					throw new XRegistryClientException(e.getLocalizedMessage(),e);
				}
			}
		 }
	    
	    /**
		 * Constructs a XRegistryClient with SSL Certificate and trsted ca's Files for secure XRegistry Operations .
		 * @param gsicredential
		 * @param trusted certificates
		 * @param secured-xregistry-url
		 * 
		 */
	    public XRegistryClient(String hostkeyPath,  String trustedCerts,
		           String xregistryServiceWsdlUrl) throws XRegistryClientException{
	    	this(hostkeyPath, CertificateManager.getTrustedCertificates(trustedCerts), xregistryServiceWsdlUrl);
		 }

	    public String[]  app2Hosts(String appName) throws XRegistryClientException {
	        App2HostsDocument document = App2HostsDocument.Factory.newInstance();
	        document.addNewApp2Hosts().setAppName(appName);
	        App2HostsResponseDocument responseDocument = xregistryStub.app2Hosts(document);
	        return responseDocument.getApp2HostsResponse().getHostArray();
	    }
	    
	    public void addCapability(String resource, String actor, boolean isUser, String action) throws XRegistryClientException {
	        AddCapabilityDocument document = AddCapabilityDocument.Factory.newInstance();
	        AddCapability capability = document.addNewAddCapability();
	        capability.setResourceID(resource);
	        capability.setActor(actor);
	        capability.setIsActorAUser(isUser);
	        capability.setAction(action);
	        xregistryStub.addCapability(document);
	    }
	    
	    /**
	     * This is the search method for capabilities, the null values can be defined for most values to allow a search. But one of the 
	     * resourceID or actor must present
	     * @param user
	     * @param resourceID
	     * @param actor and actorType - Both or non must precent .. can be Null
	     * @param action 
	     * @return
	     * @throws XRegistryClientException
	     */
	    
	    public CapabilityToken[] findCapability(String resourceID, String actor, boolean actorType, String action) throws XRegistryClientException {
	        GetCapabilityDocument document = GetCapabilityDocument.Factory.newInstance();
	        GetCapability getCapability = document.addNewGetCapability();
	        if(resourceID != null || actor != null){
	            getCapability.setUser(actor);
	            getCapability.setResourceID(resourceID);
	            getCapability.setActorType(actorType);
	            getCapability.setAction(action);
	            GetCapabilityResponseDocument capabilityResponse = xregistryStub.getCapability(document);
	            return capabilityResponse.getGetCapabilityResponse().getTokenArray();
	        }else{
	            throw new XRegistryClientException("one of the resourceID or actor must present");
	        }
	    }    
	    
	    public void removeCapability(String resourceID,String actor) throws XRegistryClientException{
	        RemoveCapabilityDocument document = RemoveCapabilityDocument.Factory.newInstance();
	        RemoveCapability removeCapability = document.addNewRemoveCapability();
	        removeCapability.setResourceID(resourceID);
	        removeCapability.setActor(actor);
	        xregistryStub.removeCapability(document);
	    }
	    
	    
	    public AppData[] findAppDesc(String query) throws XRegistryClientException {
	        FindAppDescDocument document = FindAppDescDocument.Factory.newInstance();
	        document.addNewFindAppDesc().setAppNameSearchStr(query);
	        FindAppDescResponseDocument responseDocument = xregistryStub.findAppDesc(document);
	        AppData[] results = responseDocument.getFindAppDescResponse().getAppDataArray();
	        if(results == null){
	            return null;
	        }
	        return results;
	  
	    }
	    
	    public HostDescData[]  findHosts(String hostName) throws XRegistryClientException {
	        FindHostsDocument document = FindHostsDocument.Factory.newInstance();
	        document.addNewFindHosts().setHostNameSearchStr(hostName);
	        FindHostsResponseDocument responseDocument = xregistryStub.findHosts(document);
	        
	        HostDescData[] hostDescData =  responseDocument.getFindHostsResponse().getResultsArray();
	        if(hostDescData == null){
	            return null;
	        }
	       return hostDescData;
	    }
	    
	    public ResourceData[]  findResource(String resourceNameQuery) throws XRegistryClientException {
	        FindResourceDocument document = FindResourceDocument.Factory.newInstance();
	        document.addNewFindResource().setResourceNameQuery(resourceNameQuery);
	        FindResourceResponseDocument responseDocument = xregistryStub.findResource(document);
	        
	        ResourceData[] resourceData =  responseDocument.getFindResourceResponse().getResourceArray();
	        if(resourceData == null){
	            return null;
	        }
	        return resourceData;
	    }
	    
	    public ServiceDescData[]  findServiceDesc(String serviceName) throws XRegistryClientException {
	        FindServiceDescDocument document = FindServiceDescDocument.Factory.newInstance();
	        document.addNewFindServiceDesc().setServiceQNameSearchStr(serviceName);
	        FindServiceDescResponseDocument responseDocument = xregistryStub.findServiceDesc(document);
	        
	        ServiceDescData[] serviceDescData =  responseDocument.getFindServiceDescResponse().getServiceNameArray();
	        
	        if(serviceDescData == null){
	            return null;
	        }
	       return serviceDescData;
	    }	    
	    
	    public WsdlData[]  findServiceInstance(String serviceName) throws XRegistryClientException {
	        FindServiceInstanceDocument document = FindServiceInstanceDocument.Factory.newInstance();
	        document.addNewFindServiceInstance().setServiceQnameSearchStr(serviceName);
	        FindServiceInstanceResponseDocument responseDocument = xregistryStub.findServiceInstance(document);
	        
	        WsdlData[] wsdlDescData =  responseDocument.getFindServiceInstanceResponse().getServiceInstanceArray();
	        if(wsdlDescData == null){
	            return null;
	        }
	       return wsdlDescData;
	    }
	    
	    public String getAbstractWsdl(QName wsdlQName) throws XRegistryClientException {
	        GetAbstractWsdlDocument document = GetAbstractWsdlDocument.Factory.newInstance();
	        document.addNewGetAbstractWsdl().setWsdlQName(wsdlQName);
	        GetAbstractWsdlResponseDocument responseDocument = xregistryStub.getAbstractWsdl(document);
	        return responseDocument.getGetAbstractWsdlResponse().getWsdlAsStr();
	    }
	    
	    public String getAppDesc(String appName, String hostName) throws XRegistryClientException {
	        GetAppDescDocument document = GetAppDescDocument.Factory.newInstance();
	        GetAppDesc getAppDesc = document.addNewGetAppDesc();
	        getAppDesc.setAppQName(appName);
	        getAppDesc.setHostName(hostName);
	        return xregistryStub.getAppDesc(document).getGetAppDescResponse().getAppdescAsStr();
	    }
	    
	    public String getConcreateWsdl(QName wsdlQName) throws XRegistryClientException {
	        GetConcreateWsdlDocument document = GetConcreateWsdlDocument.Factory.newInstance();
	        document.addNewGetConcreateWsdl().setWsdlQname(wsdlQName);
	        return xregistryStub.getConcreateWsdl(document).getGetConcreateWsdlResponse().getWsdlAsStr();
	    }
	    
	    public String getHostDesc(String hostName) throws XRegistryClientException {
	        GetHostDescDocument document = GetHostDescDocument.Factory.newInstance();
	        document.addNewGetHostDesc().setHostName(hostName);
	        return xregistryStub.getHostDesc(document).getGetHostDescResponse().getHostDescAsStr();
	    }
	    
	    public String getResource(QName resourceName) throws XRegistryClientException {
	        GetResourceDocument document = GetResourceDocument.Factory.newInstance();
	        document.addNewGetResource().setResourceName(resourceName);
	        return xregistryStub.getResource(document).getGetResourceResponse().getResourceAsStr();
	    }
	    
	    public String getServiceDesc(QName serviceName) throws XRegistryClientException {
	        GetServiceDescDocument input =  GetServiceDescDocument.Factory.newInstance();
	        input.addNewGetServiceDesc().setServiceQname(serviceName);
	        return xregistryStub.getServiceDesc(input).getGetServiceDescResponse().getServiceDescAsStr();
	    }
	    
	    public String registerAppDesc(String appDescAsStr) throws XRegistryClientException {
	        RegisterAppDescDocument input =  RegisterAppDescDocument.Factory.newInstance();
	        input.addNewRegisterAppDesc().setAppDescAsStr(appDescAsStr);
	        return xregistryStub.registerAppDesc(input).getRegisterAppDescResponse().getResourceID();
	    }
	    
	    public String registerConcreteWsdl( String wsdlAsStr, int lifetimeAsSeconds) throws XRegistryClientException {
	        RegisterConcreteWsdlDocument input =  RegisterConcreteWsdlDocument.Factory.newInstance();
	        RegisterConcreteWsdl concreteWsdl = input.addNewRegisterConcreteWsdl();
	        concreteWsdl.setWsdlAsStr(wsdlAsStr);
	        concreteWsdl.setLifetimeAsSeconds(lifetimeAsSeconds);
	        return xregistryStub.registerConcreteWsdl(input).getRegisterConcreteWsdlResponse().getResourceID();
	    }
	    
	    public String registerHostDesc( String hostDescAsStr) throws XRegistryClientException {
	        RegisterHostDescDocument input =  RegisterHostDescDocument.Factory.newInstance();
	        input.addNewRegisterHostDesc().setHostDescAsStr(hostDescAsStr);
	        return xregistryStub.registerHostDesc(input).getRegisterHostDescResponse().getResourceID();
	    }
	    
	    public void registerResource(QName resourceName, String resourceAsStr) throws XRegistryClientException {
	        AddResourceDocument input = AddResourceDocument.Factory.newInstance();
	        AddResource resource = input.addNewAddResource();
	        resource.setResourceAsStr(resourceAsStr);
	        resource.setResourceName(resourceName);
	        xregistryStub.addResource(input);
	    }
	    
	    public String registerServiceDesc( String serviceDescAsStr, String awsdlAsStr) throws XRegistryClientException {
	        RegisterServiceDescDocument input =  RegisterServiceDescDocument.Factory.newInstance();
	        RegisterServiceDesc serviceDesc = input.addNewRegisterServiceDesc();
	        serviceDesc.setServiceDescAsStr(serviceDescAsStr);
	        serviceDesc.setAwsdlAdStr(awsdlAsStr);
	        return xregistryStub.registerServiceDesc(input).getRegisterServiceDescResponse().getResourceID();
	    }
	    
	    public void removeAppDesc(QName appName, String hostName) throws XRegistryClientException {
	        RemoveAppDescDocument input =  RemoveAppDescDocument.Factory.newInstance();
	        RemoveAppDesc removeAppDesc = input.addNewRemoveAppDesc();
	        removeAppDesc.setAppName(appName);
	        removeAppDesc.setHostName(hostName);
	        xregistryStub.removeAppDesc(input);
	        
	    }
	    
	    public void removeConcreteWsdl(QName wsdlQName) throws XRegistryClientException {
	        RemoveConcreteWsdlDocument input =  RemoveConcreteWsdlDocument.Factory.newInstance();
	        input.addNewRemoveConcreteWsdl().setWsdlQName(wsdlQName);
	        xregistryStub.removeConcreteWsdl(input);
	    }
	    
	    public void removeHostDesc(String hostName) throws XRegistryClientException {
	        RemoveHostDescDocument input =  RemoveHostDescDocument.Factory.newInstance();
	        input.addNewRemoveHostDesc().setHostName(hostName);
	        xregistryStub.removeHostDesc(input);        
	    }
	    
	    public void removeServiceDesc(QName serviceName) throws XRegistryClientException {
	        RemoveServiceDescDocument input =  RemoveServiceDescDocument.Factory.newInstance();
	        input.addNewRemoveServiceDesc().setServiceQname(serviceName);
	        xregistryStub.removeServiceDesc(input);
	    }
	    
	    
	    public void removeResource(QName resourceName) throws XRegistryClientException {
	        RemoveResourceDocument input =  RemoveResourceDocument.Factory.newInstance();
	        input.addNewRemoveResource().setResourceName(resourceName);
	        xregistryStub.removeResource(input);
	    }
	    
	    /**
	     * Give a actor, and a resource, this decides does the actor have acess to the resource. The actor must be a user and not a group
	     * @param user
	     * @param resourceID
	     * @param actor
	     * @param action
	     * @return
	     * @throws XRegistryClientException
	     */
	    
	    public boolean isAuthorizedToAcsses(String user, String resourceID, String actor, String action) throws XRegistryClientException {
	        IsAuthorizedToAcssesDocument input = IsAuthorizedToAcssesDocument.Factory.newInstance();
	        IsAuthorizedToAcsses isAuthorizedToAcsses = input.addNewIsAuthorizedToAcsses();
	        isAuthorizedToAcsses.setActor(actor);
	        isAuthorizedToAcsses.setResourceID(resourceID);
	        isAuthorizedToAcsses.setAction(action);
	        IsAuthorizedToAcssesResponseDocument responseDocument =  xregistryStub.isAuthorizedToAcsses(input);
	        return responseDocument.getIsAuthorizedToAcssesResponse().getDecision();
	    }
	    
	    /**
	     * 
	     * @param resourceName
	     * @param resourceType
	     * @param resourceDesc
	     * @throws XRegistryClientException
	     */
		public void registerOGCEResource(QName resourceID, String resourceName, String resourceType,
				String resourceDesc, String resourceDocument,
				String resourceParentTypedID, String owner)
				throws XRegistryClientException {
			AddOGCEResourceDocument input = AddOGCEResourceDocument.Factory
					.newInstance();
			AddOGCEResource resource = input.addNewAddOGCEResource();
			resource.setResourceID(resourceID);
			resource.setResourceName(resourceName);
			resource.setResourceType(resourceType);
			resource.setResourceDesc(resourceDesc);
			resource.setResourceDocument(resourceDocument);
			if (owner != null && owner != "") {
				resource.setOwner(owner);
			}
			resource.setParentTypedID(resourceParentTypedID);
			xregistryStub.addOGCEResource(input);
		}
	    /**
	     * 
	     * @param resourceName
	     * @return
	     * @throws XRegistryClientException
	     */
	    public String getOGCEResource(QName resourceID, String resourceType, String resourceParentTypedID) throws XRegistryClientException {
	        GetOGCEResourceDocument input = GetOGCEResourceDocument.Factory.newInstance();
	        GetOGCEResource document = input.addNewGetOGCEResource();
	        document.setResourceID(resourceID);
	        document.setResourceType(resourceType);
	        document.setParentTypedID(resourceParentTypedID);
	        return xregistryStub.getOGCEResource(input).getGetOGCEResourceResponse().getResourceAsStr();
	    }
	    
	    /**
	     * 
	     * @param resourceNameQuery
	     * @param resourceType
	     * @return
	     * @throws XRegistryClientException
	     */
	    public OGCEResourceData[]  findOGCEResource(String resourceNameQuery, String resourceType, String resourceParentTypedID) throws XRegistryClientException {
	        FindOGCEResourceDocument document = FindOGCEResourceDocument.Factory.newInstance();
	        FindOGCEResource  findOGCEResource =  document.addNewFindOGCEResource();
	        findOGCEResource.setResourceNameQuery(resourceNameQuery);
	        findOGCEResource.setResourceType(resourceType);
	        findOGCEResource.setResourceName(resourceNameQuery);
	        findOGCEResource.setParentTypedID(resourceParentTypedID);
	        
	        FindOGCEResourceResponseDocument responseDocument = xregistryStub.findOGCEResource(document);
	        
	        OGCEResourceData[] resourceData =  responseDocument.getFindOGCEResourceResponse().getResourceArray();
	        if(resourceData == null){
	            return null;
	        }
	        return resourceData;
	    }
	    /**
	     * 
	     * @param resourceName
	     * @throws XRegistryClientException
	     */
	    public void removeOGCEResource(QName resourceID, String resourceType) throws XRegistryClientException {
	        RemoveOGCEResourceDocument input =  RemoveOGCEResourceDocument.Factory.newInstance();
	        RemoveOGCEResource removeOGCEResource = input.addNewRemoveOGCEResource();
	        removeOGCEResource.setResourceID(resourceID);
	        removeOGCEResource.setResourceType(resourceType);
	        xregistryStub.removeOGCEResource(input);
	    }

}